@@ -1,6 +1,7 @@
 from threading import Lock
 from time import sleep, time
 from math import ceil
+from os import environ
 
 from psutil import Process, cpu_count, cpu_times
 
@@ -12,8 +13,13 @@ CPU_IDLE_POLL_DIVISOR = 20
 CPU_IDLE_POLL_MIN_INTERVAL = 0.050
 CPU_IDLE_POLL_MAX_INTERVAL = 2
 
-# Jobs will still be clamped to idle CPU, but we want SOME kind of upper bound on concurrency, to avoid I/O
-# blockage causing ridiculous amounts of work
+# The worker will ignore available idle CPU until it is using at least this much CPU itself
+# Useful for shared environments like in Kubernetes where neighbor pods might use your `cpu.request`
+# oppotunistically, until you use it yourself
+CPU_PROMISED = int(environ.get("POTHEAD_CPU_PROMISED", 0))
+
+# Jobs will still be clamped to idle CPU, but we want SOME kind of upper bound on concurrency, to avoid
+# temporary I/O blockage causing ridiculous amounts of work
 DEFAULT_MAX_CONCURRENT_MULTIPLIER = 2
 
 __all__ = ["wait_for_idle_cpus"]
@@ -72,7 +78,7 @@ class WorkerList:
         with self.lock:
             self.workers.remove(worker)
 
-    def count(self):
+    def __len__(self):
         with self.lock:
             return len(self.workers)
 
@@ -104,36 +110,29 @@ class InertialTimeDerivate:
         return self.inertial_value.update(delta / elapsed)
 
 
-"""CPU-gated wait_for_slot implementation
-
-Creates a wait_for_slot-callable that will let through at least one concurrent job, and additional jobs as
-long as `count` cpu:s are idle. It is cgroup-aware, and will not allow current process and subprocesses to
-consume more CPU than configured cgroup-limit.
-
-When a worker is spawned, but before it has received its job request, it will have the CPU amount
-specified in the `required` parameter artificially reserved for it. The `keep_cpu_reservation`
-parameter can be used for jobs that when started don't immediately use the full amount of CPU that
-they will do eventually. In order to avoid spawning lots of such jobs in a short time period before
-CPU usage has settled, the artifical CPU reservation can be kept by setting `keep_cpu_reservation`
-to True. When doing this, the job should at some later point during its request processing reset the
-CPU reservation by calling the function stored under the "RESET_CPU_RESERVATION" key in the WSGI
-request environment, like so:
-
-    http_request.environ["RESET_CPU_RESERVATION"](new_cpu_amount)
-
-The function may be called several times during the request processing, but the last call should
-normally set the amount to 0.
-"""
-
-
 def wait_for_idle_cpus(
-    required,
-    *,
-    max_concurrent=None,
-    fixed_delay=0.2,
-    inertia=0.7,
-    keep_cpu_reservation=False
+    required, *, max_concurrent=None, inertia=0.7, keep_cpu_reservation=False
 ):
+    """CPU-gated wait_for_slot implementation
+
+    Creates a wait_for_slot-callable that will let through at least one concurrent job, and additional jobs as
+    long as `required` cpu:s are idle. It is cgroup-aware, and will not allow current process and subprocesses to
+    consume more CPU than configured cgroup-limit.
+
+    When a worker is spawned, but before it has received its job request, it will have the CPU amount
+    specified in the `required` parameter artificially reserved for it. The `keep_cpu_reservation`
+    parameter can be used for jobs that when started don't immediately use the full amount of CPU that
+    they will do eventually. In order to avoid spawning lots of such jobs in a short time period before
+    CPU usage has settled, the artifical CPU reservation can be kept by setting `keep_cpu_reservation`
+    to True. When doing this, the job should at some later point during its request processing reset the
+    CPU reservation by calling the function stored under the "RESET_CPU_RESERVATION" key in the WSGI
+    request environment, like so:
+
+        http_request.environ["RESET_CPU_RESERVATION"](new_cpu_amount)
+
+    The function may be called several times during the request processing, but the last call should
+    normally set the amount to 0.
+    """
     worker_list = WorkerList()
     cpu_wait_time = MovingAverage(1, 0.7)
 
@@ -165,17 +164,23 @@ def wait_for_idle_cpus(
                     cpu_count() * DEFAULT_MAX_CONCURRENT_MULTIPLIER / required
                 )
 
+    def cpu_is_available(idle, cpu_used):
+        reserved = worker_list.reserved_cpu_amount()
+        cpu_used += reserved
+
+        if cpu_used < CPU_PROMISED:
+            return True
+
+        if idle - reserved < required:
+            return False
+
+        return cpu_quota is None or cpu_used <= cpu_quota
+
     def wait_for_slot(halt):
-        if worker_list.count() == 0:
-            return worker_list.add_worker(Worker(required, keep_cpu_reservation))
-
-        sleep(fixed_delay)
-
         start_time = time()
 
         idle = InertialTimeDerivate(lambda: cpu_times().idle, 0, inertia)
-        if cpu_quota:
-            cpu_used = InertialTimeDerivate(cpu_used_in_process, cpu_count(), inertia)
+        cpu_used = InertialTimeDerivate(cpu_used_in_process, cpu_count(), inertia)
 
         cpu_poll_interval = max(
             CPU_IDLE_POLL_MIN_INTERVAL,
@@ -183,19 +188,20 @@ def wait_for_idle_cpus(
                 cpu_wait_time.value / CPU_IDLE_POLL_DIVISOR, CPU_IDLE_POLL_MAX_INTERVAL
             ),
         )
-        while worker_list.count() > 0 and not halt():
+        while len(worker_list) and not halt():
             reserved = worker_list.reserved_cpu_amount()
             idle_deriv = idle.update()
             if (
-                worker_list.count() < max_concurrent
+                len(worker_list) < max_concurrent
                 and idle_deriv - reserved >= required
                 and (cpu_quota is None or cpu_used.update() + reserved <= cpu_quota)
             ):
-                cpu_wait_time.update(time() - start_time)
                 break
 
             sleep(cpu_poll_interval)
 
+        cpu_wait_time.update(time() - start_time)
+
         return worker_list.add_worker(Worker(required, keep_cpu_reservation))
 
     return wait_for_slot
